/*
 * Whitelist Tool
 *
 * By Per Øyvind Karlsen <pkarlsen@mandriva.com> 2006
 * Based on s2ram from suspend by Pavel Machek <pavel@suse.cz> 2006
 *
 * Distributed under GPLv2.
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#define S2RAM
#include "vbetool/vbetool.h"
#include "vbetool/include/lrmi.h"
#include "vt.h"
#include "s2ram.h"

#define VBESTATE "/var/run/pm/vbestate"

static void *vbe_buffer;
/* Flags set from whitelist */
static int flags, vbe_mode = -1;
char bios_version[1024], sys_vendor[1024], sys_product[1024], sys_version[1024];

/* return codes for s2ram_prepare */
#define S2RAM_OK	0
#define S2RAM_FAIL	1
#define S2RAM_NOFB	126
#define S2RAM_UNKNOWN	127

/* flags for the whitelist */
#define S3_BIOS     0x01	/* machine needs acpi_sleep=s3_bios */
#define S3_MODE     0x02	/* machine needs acpi_sleep=s3_mode */
#define VBE_SAVE    0x04	/* machine needs "vbetool save / restore" */
#define VBE_POST    0x08	/* machine needs "vbetool post" */
#define RADEON_OFF  0x10	/* machine needs "radeontool light off" */
#define UNSURE      0x20	/* unverified entries from acpi-support 0.59 */
#define NOFB        0x40	/* must not use a frame buffer */
#define VBE_MODE    0x80	/* machine needs "vbetool vbemode get / set" */

#include "whitelist.c"

static void identify_machine(void)
{
	printf("This machine can be identified by:\n");
	printf("    sys_vendor   = \"%s\"\n"
	       "    sys_product  = \"%s\"\n"
	       "    sys_version  = \"%s\"\n"
	       "    bios_version = \"%s\"\n",
	       sys_vendor, sys_product, sys_version, bios_version);
	printf("See http://qa.mandriva.com/twiki/bin/view/Main/SuspendWhitelist for details.\n"
	       "\n"
	       "If you report a problem, please include the complete output "
	       "above.\n");
}

void machine_known(int i)
{
	if (i < 0) {
		printf("Internal error: machine_known, i<0. Please report.\n");
		return;
	}

	printf("Machine matched entry %d:\n"
	       "    sys_vendor   = '%s'\n"
	       "    sys_product  = '%s'\n"
	       "    sys_version  = '%s'\n"
	       "    bios_version = '%s'\n", i,
	       whitelist[i].sys_vendor, whitelist[i].sys_product,
	       whitelist[i].sys_version, whitelist[i].bios_version);
	printf("Fixes: 0x%x  %s%s%s%s%s%s%s\n", flags,
	       (flags & VBE_SAVE) ? "VBE_SAVE " : "",
	       (flags & VBE_POST) ? "VBE_POST " : "",
	       (flags & VBE_MODE) ? "VBE_MODE " : "",
	       (flags & RADEON_OFF) ? "RADEON_OFF " : "",
	       (flags & S3_BIOS) ? "S3_BIOS " : "",
	       (flags & S3_MODE) ? "S3_MODE " : "",
	       (flags & NOFB) ? "NOFB " : "");
	if (flags & UNSURE)
		printf("Machine is in the whitelist but perhaps using "
		       "vbetool unnecessarily.\n"
		       "Please try to find minimal options.\n\n");
	/* in case of a bugreport we might need to find a better match than
	 * the one we already have (additional BIOS version e.g)...
	 */
	identify_machine();
}

static int set_acpi_video_mode(int mode)
{
	FILE *f = fopen("/proc/sys/kernel/acpi_video_flags", "w");
	if (!f) {
		printf("/proc/sys/kernel/acpi_video_flags does not exist; you need a kernel >=2.6.16.\n");
		return S2RAM_FAIL;
	}
	fprintf(f, "%d", mode);
	fflush(f);
	fclose(f);
	return S2RAM_OK;
}

static int match(const char *t, const char *s)
{
	int len = strlen(s);
	/* empty string matches always */
	if (len == 0)
		return 1;

	if (s[len-1] == '*') {
		len--;
		return !strncmp(t, s, len);
	} else {
		return !strcmp(t,s);
	}
}

static int machine_match(void)
{
	int i;
	/* sys_vendor = NULL terminates the whitelist array */
	for (i = 0; whitelist[i].sys_vendor; i++) {
		if (match(sys_vendor,   whitelist[i].sys_vendor)  &&
		    match(sys_product,  whitelist[i].sys_product) &&
		    match(sys_version,  whitelist[i].sys_version) &&
		    match(bios_version, whitelist[i].bios_version)) {
			return i;
		}
	}
	return -1;
}

int s2ram_check(int id)
{
	int ret = S2RAM_OK;

	if (id < 0) {
		ret = S2RAM_UNKNOWN;
	} else {
		flags = whitelist[id].flags;
		if ((flags & NOFB) && is_framebuffer())
			ret = S2RAM_NOFB;
	}

	return ret;
}

/* warning: we have to be on a text console when calling this */
int s2ram_hacks(void)
{
	int ret = 0;

	/* 0 means: don't touch what was set on kernel commandline */
	if (flags & (S3_BIOS | S3_MODE))
		ret = set_acpi_video_mode(flags & (S3_BIOS | S3_MODE));

	if (ret)
		return ret;

	return 0;
}

int s2ram_prepare(void)
{
	int ret, id;

	dmi_scan();
	id = machine_match();
	ret = s2ram_check(id);

	if (ret)
		return ret;

	ret = s2ram_hacks();

	return ret;
}


int s2ram_do(void)
{
	if (flags & VBE_SAVE) {
		int size;
		int fil;
		vbetool_init();
		printf("Calling save_state using" VBESTATE "\n");
		vbe_buffer = __save_state(&size);
		if(vbe_buffer){
			fil = open(VBESTATE,O_WRONLY|O_CREAT,0600);
			if(write(fil, vbe_buffer, size) == -1)
				printf("Unable to write to " VBESTATE "\n");
			close(fil);
		}
	}
	if (flags & VBE_MODE) {
		vbetool_init();
		printf("Calling get_mode\n");
		vbe_mode = __get_mode();
	}
	if (flags & RADEON_OFF) {
		map_radeon_cntl_mem();
		printf("Calling radeon_cmd_light(0)\n");
		radeon_cmd_light(0);
	}

	return 0;
}

void s2ram_resume(void)
{
	int fil;
	char *data = NULL;
	struct stat buffer;

	// FIXME: can we call vbetool_init() multiple times without cleaning up?
	if (flags & VBE_POST) {
		vbetool_init();
		printf("Calling do_post\n");
		do_post();
	}
	if (flags & VBE_SAVE) {
		vbetool_init();
		fil = open(VBESTATE,O_RDONLY,0600);
		stat(VBESTATE,&buffer);
		data = LRMI_alloc_real(buffer.st_size);
		if(read(fil, data, buffer.st_size) == -1)
			printf("Unable to read " VBESTATE "\n");
		close(fil);
		printf("Calling restore_state_from using " VBESTATE "\n");
		restore_state_from(data);
	}
	if (flags & VBE_MODE) {
		vbetool_init();
		printf("Calling set_vbe_mode\n");
		set_vbe_mode(vbe_mode);
	}
	if (flags & RADEON_OFF) {
		printf("Calling radeon_cmd_light(1)\n");
		radeon_cmd_light(1);
	}
}

static void usage(void)
{
	printf("Usage: wltool [-nhiq] [-fa] [-fSasrm] [-fRasprm]\n"
	       "\n"
	       "Options:\n"
	       "    -h, --help:       this text.\n"
	       "    -n, --test:       test if the machine is in the database.\n"
	       "                      returns 0 if known and supported\n"
	       "    -i, --identify:   prints a string that identifies the machine.\n"
	       "    -S, --suspend:    prepare for suspend.\n"
	       "    -R, --resume:     post resume fixes .\n"
	       "    -q, --quiet:      be quiet even if machine is unsupported or\n"
	       "                      framebuffer cannot be used.\n"
	       "                      (nice to avoid pollution of boot messages)\n"
	       "    -f, --force:      force options, even on unknown machines.\n"
	       "\n"
	       "the following options are only available with --force:\n"
	       "    -a, --acpi_sleep: set the acpi_sleep parameter\n"
	       "                      1=s3_bios, 2=s3_mode, 3=both\n"
	       "the following options are only available with --suspend in addition to --force:\n"
	       "    -s, --vbe_save:   save VBE state.\n"
	       "    -m, --vbe_mode:   get VBE mode.\n"
	       "    -r, --radeontool: turn off the backlight on radeons.\n"
	       "the following options are only available with --resume in addition to --force:\n"
	       "    -s, --vbe_save:   restore VBE state.\n"
	       "    -p, --vbe_post:   VBE POST the graphics card after resume.\n"
	       "    -m, --vbe_mode:   set VBE mode.\n"
	       "    -r, --radeontool: turn on the backlight on radeons.\n"
	       "\n");
	exit(1);
}

int main(int argc, char *argv[])
{
	int i, id = -1, ret = 0, test_mode = 0, force = 0, quiet = 0, suspend = 0, resume = 0;
	struct option options[] = {
		{ "test",	no_argument,		NULL, 'n'},
		{ "help",	no_argument,		NULL, 'h'},
		{ "force",	no_argument,		NULL, 'f'},
		{ "vbe_save",	no_argument,		NULL, 's'},
		{ "vbe_post",	no_argument,		NULL, 'p'},
		{ "vbe_mode",	no_argument,		NULL, 'm'},
		{ "radeontool",	no_argument,		NULL, 'r'},
		{ "identify",	no_argument,		NULL, 'i'},
		{ "acpi_sleep",	required_argument,	NULL, 'a'},
		{ "suspend", 	no_argument,		NULL, 'S'},
		{ "resume",	no_argument,		NULL, 'R'},
		{ "quiet",	no_argument,		NULL, 'q'},
		{ NULL,		0,			NULL,  0 }
	};

	while ((i = getopt_long(argc, argv, "nhfspmriRSqa:", options, NULL)) != -1) {
		switch (i) {
		case 'h':
			usage();
			break;
		case 'i':
			dmi_scan();
			identify_machine();
			exit(0);
		case 'n':
			test_mode = 1;
			break;
		case 'f':
			force = 1;
			break;
                case 's':
			flags |= VBE_SAVE;
			break;
		case 'p':
			flags |= VBE_POST;
			break;
		case 'm':
			flags |= VBE_MODE;
			break;
		case 'r':
			flags |= RADEON_OFF;
			break;
		case 'a':
			flags |= (atoi(optarg) & (S3_BIOS | S3_MODE));
			break;
		case 'S':
			suspend = 1;
			break;
		case 'R':
			resume = 1;
			break;
		case 'q':
			quiet = 1;
			break;
		default:
			usage();
			break;
		}
	}
	if (flags && !force) {
		printf("acpi_sleep, vbe_save, vbe_post and radeontool parameter "
		       "must be used with --force\n\n");
		usage();
	}

	if (suspend && resume) {
		printf("You cannot suspend & resume at the same time!\n\n");
		usage();
	}

        if (flags & (VBE_SAVE)  && (!suspend && !resume)){
		printf("--vbe_save can only be used with --suspend or --resume!\n\n");
		usage();
	}

	if (flags & (VBE_POST) && !resume){
		printf("--vbe_post can only be used with --resume!\n\n");
		usage();
	}

	if (flags & (VBE_MODE) && (!suspend && !resume)){
		printf("--vbe_mode can only be used with --suspend or --resume\n\n");
		usage();
	}

	if (flags & (RADEON_OFF) && (!suspend && !resume)){
		printf("--radeontool can only be used with --suspend or --resume\n\n");
		usage();
	}

	if (force && test_mode) {
		printf("force and test mode do not make sense together.\n\n");
		usage();
	}

	if (!force) {
		dmi_scan();
		id = machine_match();
		ret = s2ram_check(id);
	}

	if (ret == S2RAM_UNKNOWN && !quiet) {
		printf("Machine is unknown.\n");
		identify_machine();
		goto out;
	}

	if (ret == S2RAM_NOFB && !quiet)
		printf("This machine can only suspend without framebuffer.\n");

	if (test_mode) {
		machine_known(id);
		goto out;
	}

	if (ret)
		goto out;

	ret = s2ram_hacks();
	if (ret)
		goto out;
	if(suspend)
		s2ram_do();
	if(resume)
		s2ram_resume();

 out:
	return ret;
}
